package org.neuroph.contrib.tileclassification.model;

import java.awt.Dimension;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.Map.Entry;

import org.neuroph.contrib.tileclassification.model.labels.BinaryLabelResponses;
import org.neuroph.contrib.tileclassification.model.labels.LabelResponses;
import org.neuroph.contrib.tileclassification.model.plugin.TileClassificationPlugin;
import org.neuroph.contrib.tileclassification.screentools.FractionRgbData;
import org.neuroph.core.NeuralNetwork;
import org.neuroph.core.Neuron;
import org.neuroph.core.learning.SupervisedTrainingElement;
import org.neuroph.core.learning.TrainingSet;
import org.neuroph.nnet.MultiLayerPerceptron;
import org.neuroph.nnet.learning.BackPropagation;
import org.neuroph.nnet.learning.MomentumBackpropagation;
import org.neuroph.util.VectorParser;
import org.neuroph.util.plugins.LabelsPlugin;
import org.neuroph.util.plugins.PluginBase;

public class NeuralNetworkManager {
	public static final String DEFAULT_ANN_FILE_EXTENSION = "ann";
	
	private NeuralNetwork activeNeuralNetwork;
	
	public NeuralNetwork getActiveNeuralNetwork() {
		return activeNeuralNetwork;
	}
	
	public void loadNeuralNetwork(File dir, String neuralNetworkLabel) 
		throws IOException, ClassNotFoundException
	{
		if(!dir.isDirectory()) {
			throw new IOException("The specified file object must be a directory!");
		}
		
		File annFile = new File(dir, neuralNetworkLabel + "." + DEFAULT_ANN_FILE_EXTENSION);
		if(annFile.exists()) {
			InputStream inputStream = new FileInputStream(annFile);
			
			try {
				ObjectInputStream in = new ObjectInputStream(inputStream);
				activeNeuralNetwork = (NeuralNetwork) in.readObject();
			} finally {
				if(inputStream != null) {
					inputStream.close();
				}
			}
		} else {
			throw new FileNotFoundException("No neural network exists at: " + 
					annFile.getAbsolutePath());
		}
	}
	
	public File persistNeuralNetwork(File dir) throws IOException {
		if(activeNeuralNetwork == null) {
			throw new IOException("Cannot persist the active neural network because it is null.");
		}
		
		if(!dir.isDirectory()) {
			throw new IOException("The specified file object must be a directory!");
		}
		
		ObjectOutput out = null;
		try {
			if(!dir.exists()) {
				dir.mkdirs();
			}
			
			LabelsPlugin labelsPlugin = (LabelsPlugin) activeNeuralNetwork.getPlugin(LabelsPlugin.LABELS_PLUGIN_NAME);
			String annName = labelsPlugin.getLabel(activeNeuralNetwork);
			
			File destination = new File(dir, annName + "." + DEFAULT_ANN_FILE_EXTENSION);
			out = new ObjectOutputStream(new FileOutputStream(destination));
			out.writeObject(activeNeuralNetwork);
			return destination;
		} finally {
			try {
				if(out != null) {
					out.close();
				}
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}
	
	/**
	 * Assumes that all of the FractionRgbData objects in the given map have identical 
	 * length arrays in them so that the input layer of the neural network can be 
	 * created here.
	 * 
	 * @param targetLabels
	 * @param rgbDataMap 
	 */
	public void createNewActiveClassificationNeuralNetwork(String neuralNetworkName, Dimension samplingResolution, LabelResponses labelResponses, boolean useMomentumBackprop, List<String> targetLabels, Map<String, FractionRgbData> rgbDataMap) {
		int numberOfInputNeurons = 3 * samplingResolution.width * samplingResolution.height;
		int numberOfOuputNeurons = targetLabels.size();

		int numberOfHiddenNeurons = targetLabels.size();
		if(numberOfHiddenNeurons < 12) {
			numberOfHiddenNeurons = 12;
		} else if(numberOfHiddenNeurons > 128) {
			numberOfHiddenNeurons = 128;
		}
		
		Vector<Integer> layerNeuronCounts = new Vector<Integer>();
		layerNeuronCounts.add(numberOfInputNeurons);
		layerNeuronCounts.add(numberOfHiddenNeurons);
		if(!(labelResponses instanceof BinaryLabelResponses)) {
			int analogLayerCount = numberOfOuputNeurons * 2;
			if(analogLayerCount < 4) {
				analogLayerCount = 4;
			}
			layerNeuronCounts.add(analogLayerCount); //this extra layer helps with analog networks
		}
		layerNeuronCounts.add(numberOfOuputNeurons);
		
		System.out.println("Neuron layer size counts vector = " + layerNeuronCounts);
		
		activeNeuralNetwork = new MultiLayerPerceptron(layerNeuronCounts);
		PluginBase tileClassificationPlugin = new TileClassificationPlugin(samplingResolution, labelResponses);
		activeNeuralNetwork.addPlugin(tileClassificationPlugin);
		LabelsPlugin labelsPlugin = (LabelsPlugin) activeNeuralNetwork.getPlugin(LabelsPlugin.LABELS_PLUGIN_NAME);
		labelsPlugin.setLabel(activeNeuralNetwork, neuralNetworkName);
		assignLabelsToOutputNeurons(targetLabels);
		createAndSetTrainingSet(useMomentumBackprop, targetLabels, rgbDataMap);
	}
	
	private void assignLabelsToOutputNeurons(List<String> targetLabels) {
		Vector<Neuron> outputNeurons = activeNeuralNetwork.getOutputNeurons();
		LabelsPlugin labelsPlugin = (LabelsPlugin) activeNeuralNetwork.getPlugin(LabelsPlugin.LABELS_PLUGIN_NAME);
		
		for(int i=0; i<outputNeurons.size(); i++) {
			Neuron neuron = outputNeurons.get(i);
			String label = targetLabels.get(i);
			labelsPlugin.setLabel(neuron, label);
		}
	}

	public void createAndSetTrainingSet(boolean useMomentumBackprop, List<String> targetLabels, Map<String, FractionRgbData> rgbDataMap)
	{
		LabelResponses labelResponses;
		TileClassificationPlugin tileClassificationPlugin = (TileClassificationPlugin) activeNeuralNetwork.getPlugin(TileClassificationPlugin.PLUGIN_NAME);
		if(tileClassificationPlugin == null) {
			System.err.println("TileClassificationPlugin was null!  Defaulting to BinaryLabelResponses.");
			labelResponses = new BinaryLabelResponses();
		} else {
			labelResponses = tileClassificationPlugin.getLabelResponses();
		}
		
		TrainingSet trainingSet = new TrainingSet();

		for (Entry<String, FractionRgbData> entry : rgbDataMap.entrySet()) {
			double[] input = entry.getValue().getFlattenedRgbValues();
			double[] response = labelResponses.createResponse(entry.getKey(), targetLabels);
			trainingSet.addElement(new SupervisedTrainingElement(
					VectorParser.convertToVector(input), 
					VectorParser.convertToVector(response)));
		}
		
		if(useMomentumBackprop) {
			activeNeuralNetwork.setLearningRule(new MomentumBackpropagation(activeNeuralNetwork));
		} else {
			activeNeuralNetwork.setLearningRule(new BackPropagation(activeNeuralNetwork));
		}
		activeNeuralNetwork.getLearningRule().setTrainingSet(trainingSet);
	}
	
}
